import unittest
from asq.queryables import Queryable
from helpers import infinite

__author__ = "Sixty North"

class TestContains(unittest.TestCase):

    def test_contains_positive(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).contains(2)
        self.assertTrue(b)

    def test_contains_negative(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).contains(6)
        self.assertFalse(b)

    def test_contains_infinite(self):
        a = infinite()
        b = Queryable(a).contains(37)
        self.assertTrue(b)

    def test_contains_non_sequence_positive(self):
        def series():
            yield 5
            yield 7
            yield -3
            yield 2
        a = series()
        b = Queryable(a).contains(2)
        self.assertTrue(b)

    def test_contains_non_sequence_negative(self):
        def series():
            yield 5
            yield 7
            yield -3
            yield 2
        a = series()
        b = Queryable(a).contains(6)
        self.assertFalse(b)

    def test_contains_empty(self):
        b = Queryable([]).contains(7)
        self.assertFalse(b)

    def test_contains_comparer1(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).contains(15, lambda a, b: abs(a) == abs(b))
        self.assertTrue(b)

    def test_contains_comparer2(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).contains(-9, lambda a, b: abs(a) == abs(b))
        self.assertTrue(b)

    def test_contains_comparer3(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).contains(4, lambda a, b: abs(a) == abs(b))
        self.assertFalse(b)

    def test_contains_non_callable_comparator(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        self.assertRaises(TypeError, lambda: Queryable(a).contains(-9, "not callable"))

    def test_contains_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.contains(5))





import unittest
from asq.queryables import Queryable
from helpers import infinite, TracingGenerator

class TestSelectMany(unittest.TestCase):

    def test_select_many_projector_finite(self):
        a = ['fox', 'kangaroo', 'bison', 'bear']
        b = Queryable(a).select_many(lambda x : x).to_list()
        c = ['f', 'o', 'x', 'k', 'a', 'n', 'g', 'a', 'r', 'o', 'o', 'b', 'i', 's', 'o', 'n', 'b', 'e', 'a', 'r']
        self.assertEqual(b, c)

    def test_select_many_projector_selector_finite(self):
        a = ['fox', 'kangaroo', 'bison', 'bear']
        b = Queryable(a).select_many(lambda x : x, lambda y: chr(ord(y)+1)).to_list()
        c = ['g', 'p', 'y', 'l', 'b', 'o', 'h', 'b', 's', 'p', 'p', 'c', 'j', 't', 'p', 'o', 'c', 'f', 'b', 's']
        self.assertEqual(b, c)

    def test_select_many_projector_not_callable(self):
        a = ['fox', 'kangaroo', 'bison', 'bear']
        self.assertRaises(TypeError, lambda: Queryable(a).select_many("not callable", lambda y: chr(ord(y)+1)))

    def test_select_many_infinite(self):
        a = infinite()
        b = Queryable(a).select_many(lambda x: [x] * x).take(10).to_list()
        c = [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]
        self.assertEqual(b, c)

    def test_select_many_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = Queryable(a).select_many(lambda x: [x] * x)
        self.assertEqual(a.trace, [])
        b.take(10).to_list()
        self.assertEqual(a.trace, [0, 1, 2, 3, 4])

    def test_select_many_selector_not_callable(self):
        a = ['fox', 'kangaroo', 'bison', 'bear']
        self.assertRaises(TypeError, lambda: Queryable(a).select_many(lambda x : x, "not callable"))

    def test_select_many_closed(self):
        b = Queryable([1, 2, 4, 8])
        b.close()
        self.assertRaises(ValueError, lambda: b.select_many(lambda x: x * 2))


import unittest
from asq.initiators import repeat

__author__ = "Sixty North"

class TestRepeat(unittest.TestCase):

    def test_repeat(self):
        b = repeat('x', 3).to_list()
        self.assertEqual(b, ['x', 'x', 'x'])

    def test_repeat_count_zero(self):
        b = repeat('y', 0).to_list()
        self.assertEqual(b, [])

    def test_repeat_negative_count(self):
        self.assertRaises(ValueError, lambda: repeat(0, -1))





'''Predicate function factories'''

__author__ = 'Sixty North'


def eq_(rhs):
    '''Create a predicate which tests its argument for equality with a value.

    Args:
        rhs: (right-hand-side) The value with which the left-hand-side element
            will be compared for equality.

    Returns:
        A unary predicate function which compares its single argument (lhs)
        for equality with rhs.
    '''
    return lambda lhs: lhs == rhs


def ne_(rhs):
    '''Create a predicate which tests its argument for inequality with a value.

    Args:
        rhs: (right-hand-side) The value with which the left-hand-side element
            will be compared for inequality.

    Returns:
        A unary predicate function which compares its single argument (lhs)
        for inequality with rhs.
    '''
    return lambda lhs: lhs != rhs


def lt_(rhs):
    '''Create a predicate which performs a less-than comparison of its argument
    with a value.

    Args:
        rhs: (right-hand-side) The value against which the less-than test will
            be performed.

    Returns:
        A unary predicate function which determines whether its single
        argument (lhs) is less-than rhs.
    '''
    return lambda lhs: lhs < rhs


def le_(rhs):
    '''Create a predicate which performs a less-than-or-equal comparison of its
    argument with a value.

    Args:
        rhs: (right-hand-side) The value against which the less-than-or-equal
            test will be performed.

    Returns:
        A unary predicate function which determines whether its single
        argument (lhs) is less-than-or-equal to rhs.
    '''
    return lambda lhs: lhs <= rhs


def ge_(rhs):
    '''Create a predicate which performs a greater-than-or-equal comparison of
    its argument with a value.

    Args:
        rhs: (right-hand-side) The value against which the greater-than-or-
            equal test will be performed.

    Returns:
        A unary predicate function which determines whether its single
        argument (lhs) is greater-than rhs.
    '''
    return lambda lhs: lhs >= rhs


def gt_(rhs):
    '''Create a predicate which performs a greater-than comparison of its
    argument with a value.

    Args:
        rhs: (right-hand-side) The value against which the greater-than test
            will be performed.

    Returns:
        A unary predicate function which determines whether its single
        argument (lhs) is less-than-or-equal to rhs.
    '''
    return lambda lhs: lhs > rhs


def is_(rhs):
    '''Create a predicate which performs an identity comparison of its
    argument with a value.

    Args:
        rhs: (right-hand-side) The value against which the identity test will
            be performed.

    Returns:
        A unary predicate function which determines whether its single
        arguments (lhs) has the same identity - that is, is the same object -
        as rhs.
    '''
    return lambda lhs: lhs is rhs


def contains_(lhs):
    '''Create a unary predicate which tests for membership if its argument.

    Args:
        lhs: (left-hand-side) The value to test for membership for in the
            predicate argument.

    Returns:
        A unary predicate function which determines whether its single
        arguments (lhs) contains lhs.
    '''
    return lambda rhs: lhs in rhs


def not_(predicate):
    '''A predicate combinator which negates produces an inverted predicate.

    The predicate returned by this combinator is the logical inverse of the
    supplied combinator.

    Args:
        predicate: A unary predicate function to be inverted.

    Returns:
        A unary predicate function which is the logical inverse of pred.
    '''
    return lambda lhs: not predicate(lhs)


def and_(predicate1, predicate2):
    '''A predicate combinator which produces the a new predicate which is the
    logical conjunction of two existing unary predicates.

    The predicate returned by this combinator returns True when both of the two
    supplied predicates return True, otherwise it returns False.

    Args:
        predicate1: A unary predicate function.
        predicate2: A unary predicate function.

    Returns:
        A unary predicate function which is the logical conjunction of
        predicate1 and predicate2.
    '''
    return lambda lhs: predicate1(lhs) and predicate2(lhs)


def or_(predicate1, predicate2):
    '''A predicate combinator which produces the a new predicate which is the
    logical disjunction of two existing unary predicates.

    The predicate returned by this combinator returns True when either or both
    of the two supplied predicates return True, otherwise it returns False.

    Args:
        predicate1: A unary predicate function.
        predicate2: A unary predicate function.

    Returns:
        A unary predicate function which is the logical disjunction of
        predicate1 and predicate2.
    '''
    return lambda lhs: predicate1(lhs) or predicate2(lhs)


def xor_(predicate1, predicate2):
    '''A predicate combinator which produces the a new predicate which is the
    logical exclusive disjunction of two existing unary predicates.

    The predicate returned by this combinator returns True when the two
    supplied predicates return the different values, otherwise it returns
    False.

    Args:
        predicate1: A unary predicate function.
        predicate2: A unary predicate function.

    Returns:
        A unary predicate function which is the logical exclusive disjunction
        of predicate1 and predicate2.
    '''
    return lambda lhs: predicate1(lhs) != predicate2(lhs)



import unittest
from asq.queryables import Queryable
from helpers import infinite

__author__ = "Sixty North"

class TestFirstOrDefault(unittest.TestCase):

    def test_first_or_default(self):
        a = [42, 45, 23, 12]
        b = Queryable(a).first_or_default(37)
        self.assertEqual(b, 42)

    def test_first_or_default_empty(self):
        a = []
        b = Queryable(a).first_or_default(37)
        self.assertEqual(b, 37)

    def test_first_or_default_infinite(self):
        b = Queryable(infinite()).first_or_default(37)
        self.assertEqual(b, 0)

    def test_first_or_default_predicate(self):
        a = [37, 54, 23, 12]
        b = Queryable(a).first_or_default(78, lambda x: x >= 50)
        self.assertEqual(b, 54)

    def test_first_or_default_predicate_empty(self):
        a = []
        b = Queryable([]).first_or_default(56, lambda x: x >= 50)
        self.assertEqual(b, 56)

    def test_first_predicate_missing(self):
        a = [37, 42, 23, 12]
        b = Queryable(a).first_or_default(98, lambda x: x >= 50)
        self.assertEqual(b, 98)

    def test_first_predicate_not_callable(self):
        a = [37, 42, 23, 12]
        self.assertRaises(TypeError, lambda: Queryable(a).first_or_default(98, "not callable"))

    def test_first_or_default_closed(self):
        b = Queryable([])
        b.close()
        self.assertRaises(ValueError, lambda: b.first_or_default(37))



"""
test_queryable.py Unit tests for asq.queryable.Queryable
"""
import unittest

from asq.queryables import Queryable


class TestQueryable(unittest.TestCase):

    def test_non_iterable(self):
        self.assertRaises(TypeError, lambda: Queryable(5))



import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestMin(unittest.TestCase):

    def test_min(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -8, 7]
        b = Queryable(a).min()
        self.assertEqual(b, -8)

    def test_min_selector(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -8, 7]
        b = Queryable(a).min(abs)
        self.assertEqual(b, 1)

    def test_min_selector_not_callable(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -13, 7]
        self.assertRaises(TypeError, lambda: Queryable(a).min("not callable"))

    def test_min_empty(self):
        self.assertRaises(ValueError, lambda: Queryable([]).min())

    def test_min_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.min())


import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestOfType(unittest.TestCase):

    def test_of_type_int(self):
        a = ['one', 2, 3, 'four', 'five', 6, 'seven', 8, 9, 'ten']
        b = Queryable(a).of_type(int).to_list()
        c = [2, 3, 6, 8, 9]
        self.assertEqual(b, c)

    def test_of_type_str(self):
        a = ['one', 2, 3, 'four', 'five', 6, 'seven', 8, 9, 'ten']
        d = Queryable(a).of_type(str).to_list()
        e = ['one', 'four', 'five', 'seven', 'ten']
        self.assertEqual(d, e)

    def test_of_type_tuple(self):
        a = ['one', 2, 3, 'four', 3.2, 'five', 6, 'seven', 8, 9, 'ten', 9.4]
        d = Queryable(a).of_type((str, int)).to_list()
        e = ['one', 2, 3, 'four', 'five', 6, 'seven', 8, 9, 'ten']
        self.assertEqual(d, e)

    def test_of_type_not_type(self):
        a = ['one', 2, 3, 'four', 'five', 6, 'seven', 8, 9, 'ten']
        self.assertRaises(TypeError, lambda: Queryable(a).of_type(7))

    def test_of_type_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.of_type(str))


import unittest
from asq.comparers import insensitive_eq

__author__ = "Sixty North"

class TestComparers(unittest.TestCase):

    def test_case_insensitive_positive(self):
        self.assertTrue(insensitive_eq("abcd", "ABCD"))

    def test_case_insensitive_negative(self):
        self.assertFalse(insensitive_eq("abcd", "wxyz"))




import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestAll(unittest.TestCase):

    def test_all_positive(self):
        a = [True, True, True, True]
        b = Queryable(a).all()
        self.assertTrue(b)

    def test_all_negative(self):
        a = [True, True, True, False]
        b = Queryable(a).all()
        self.assertFalse(b)

    def test_all_default_predicate_positive(self):
        a = [1, 1, 1, 1]
        b = Queryable(a).all()
        self.assertTrue(b)

    def test_all_default_predicate_negative(self):
        a = [1, 1, 1, 0]
        b = Queryable(a).all()
        self.assertFalse(b)

    def test_all_predicate_positive(self):
        a = [10, 13, 25, 14]
        b = Queryable(a).all(lambda x: x >= 10)
        self.assertTrue(b)

    def test_all_predicate_negative(self):
        a = [10, 13, 25, 8]
        b = Queryable(a).all(lambda x: x >= 10)
        self.assertFalse(b)

    def test_all_empty(self):
        a = []
        b = Queryable(a).all()
        self.assertTrue(b)

    def test_all_predicate_non_callable(self):
        a = [True, True, True]
        self.assertRaises(TypeError, lambda: Queryable(a).all("not callable"))

    def test_all_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.all())


import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class Logger(object):

    def __init__(self):
        self.log = []

    def debug(self, message):
        self.log.append(message)

class TestLog(unittest.TestCase):

    def test_log_default(self):
        a = [1, 6, 4, 3, 9, 2]
        b = Queryable(a).log().to_list()
        self.assertEqual(a, b)

    def test_log_disable(self):
        a = [1, 6, 4, 3, 9, 2]
        b = Queryable(a).log(logger=None).to_list()
        self.assertEqual(a, b)

    def test_log_enable_deferred(self):
        a = [1, 6, 4, 3, 9, 2]
        logger = Logger()
        b = Queryable(a).log(logger, eager=False).take(3).to_list()
        self.assertEqual(len(logger.log), 4)
        self.assertTrue(logger.log[0].find('BEGIN') != -1)
        self.assertTrue(logger.log[0].find('DEFERRED') != -1)
        self.assertTrue(logger.log[1].find('yields 1') != -1)
        self.assertTrue(logger.log[2].find('yields 6') != -1)
        self.assertTrue(logger.log[3].find('yields 4') != -1)

    def test_log_enable_eager(self):
        a = [1, 6, 4, 3, 9, 2]
        logger = Logger()
        b = Queryable(a).log(logger, eager=True).to_list()
        self.assertEqual(len(logger.log), 8)
        self.assertTrue(logger.log[0].find('BEGIN') != -1)
        self.assertTrue(logger.log[0].find('EAGER') != -1)
        self.assertTrue(logger.log[1].find('= 1') != -1)
        self.assertTrue(logger.log[2].find('= 6') != -1)
        self.assertTrue(logger.log[3].find('= 4') != -1)
        self.assertTrue(logger.log[4].find('= 3') != -1)
        self.assertTrue(logger.log[5].find('= 9') != -1)
        self.assertTrue(logger.log[6].find('= 2') != -1)
        self.assertTrue(logger.log[7].find('END') != -1)

    def test_log_label_deferred(self):
        a = [1, 6, 4, 3, 9, 2]
        logger = Logger()
        b = Queryable(a).log(logger, label="Test1", eager=False).take(3).to_list()
        for line in logger.log:
            self.assertTrue(line.startswith("Test1"))

    def test_log_label_eager(self):
        a = [1, 6, 4, 3, 9, 2]
        logger = Logger()
        b = Queryable(a).log(logger, label="Test2", eager=True).to_list()
        for line in logger.log:
            self.assertTrue(line.startswith("Test2"))

    def test_log_interleaved_deferred(self):
        a = [1, 6, 4, 3, 9, 2]
        logger = Logger()
        b = Queryable(a).log(logger, label="source") \
            .select(lambda x: x * 2).log(logger, label="times two") \
            .where(lambda x: x % 3).log(logger, label="mod three")  \
            .take(2).log(logger, label="take two").to_list()
        c = ['take two : BEGIN (DEFERRED)',
             'mod three : BEGIN (DEFERRED)',
             'times two : BEGIN (DEFERRED)',
             'source : BEGIN (DEFERRED)',
             'source : [0] yields 1',
             'times two : [0] yields 2',
             'mod three : [0] yields 2',
             'take two : [0] yields 2',
             'source : [1] yields 6',
             'times two : [1] yields 12',
             'source : [2] yields 4',
             'times two : [2] yields 8',
             'mod three : [1] yields 8',
             'take two : [1] yields 8',
             'take two : END (DEFERRED)']
        self.assertEqual(logger.log, c)

    def test_log_interleaved_eager(self):
        a = [1, 6, 4, 3, 9, 2]
        logger = Logger()
        b = Queryable(a).log(logger, label="source", eager=True) \
            .select(lambda x: x * 2).log(logger, label="times two", eager=True) \
            .where(lambda x: x % 3).log(logger, label="mod three", eager=True)  \
            .take(2).log(logger, label="take two", eager=True).to_list()
        c = ['source : BEGIN (EAGER)',
             'source : [0] = 1',
             'source : [1] = 6',
             'source : [2] = 4',
             'source : [3] = 3',
             'source : [4] = 9',
             'source : [5] = 2',
             'source : END (EAGER)',
             'times two : BEGIN (EAGER)',
             'times two : [0] = 2',
             'times two : [1] = 12',
             'times two : [2] = 8',
             'times two : [3] = 6',
             'times two : [4] = 18',
             'times two : [5] = 4',
             'times two : END (EAGER)',
             'mod three : BEGIN (EAGER)',
             'mod three : [0] = 2',
             'mod three : [1] = 8',
             'mod three : [2] = 4',
             'mod three : END (EAGER)',
             'take two : BEGIN (EAGER)',
             'take two : [0] = 2',
             'take two : [1] = 8',
             'take two : END (EAGER)' ]
        self.assertEqual(logger.log, c)

    def test_log_closed(self):
        a = [1, 6, 4, 3, 9, 2]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.log())



'''Selector functions and selector function factories.'''

import operator
from asq._types import is_string
from asq.initiators import query

__author__ = 'Sixty North'


a_ = operator.attrgetter
'''Create a selector function which selects an attribute by name.

The callable object returned by this function fetches one or more
attributes from its only operand using the operand's __getitem__() method.
If multiple items are specified a tuple of retrieved attribute values will
be returned.

Args:
    name: The name of the attribute which will be retrieved from each
        element. The attribute name may contain dots which will be resolved
        through sequential attribute lookups.

    *args: Optional additional attribute names which will be used as
        additional attribute names for lookup.  If supplied then the created
        selector will return a tuple of values. The attribute name may contain
        dots which will be resolved through sequential attribute lookups.

Returns:
    A unary selector function which retrieves the named attribute(s) from
    its only argument and returns the value(s) of those attribute(s).
'''


m_ = operator.methodcaller
'''Create a selector function which calls a named method.

Args:
    name: The name of the method which will be called on each element.

    *args: Any optional positional arguments which will be passed to the
        called method.

    **kwargs: Any optional named arguments which will be passed to the
        called method.

Returns:
    A unary selector function which calls the named method with any
    optional positional or named arguments and which returns the
    result of the method call.
'''

k_ = operator.itemgetter
'''Create a selector function which indexes into the element by key.

The callable object returned by this function fetches one or more items from
its only operand using the operand's __getitem__() method. If multiple items
are specified a tuple of looked-up values will be returned.

Args:
    key: The key which the generated selector will use to index into
        elements.

    *args: Optional additional arguments which will be used as additional keys
        for lookup.  If supplied then the created selector will return a tuple
        of values.

Returns:
    A unary selector function which indexes into its only argument with
    the supplied key value(s).
'''


def aq_(name):
    '''
    Args:
        name: The name of the attribute which will be retrieved from each
            element. The attribute name may contain dots which will be resolved
            through sequential attribute lookups.

    Returns:
        A unary selector function which retrieves the named attribute(s) from
        its only argument and returns the value(s) of those attribute(s) as
        a Queryable.
    '''
    return lambda item: query(getattr(item, name))


def make_selector(value):
    '''Create a selector callable from the supplied value.

    Args:
        value: If is a callable, then returned unchanged.  If a string is used
            then create an attribute selector. If in an list containing a single item
            is used then create a key/index selector.

    Returns:
        A callable selector based on the supplied value.

    Raises:
        ValueError: If a selector cannot be created from the value.
    '''
    if callable(value):
        return value
    if is_string(value):
        return a_(value)
    if isinstance(value, list) and len(value) == 1:
        return k_(value[0])
    raise ValueError("Unable to create callable selector from '{0}'".format(value))


def identity(x):
    '''The identity function.

    The identity function returns its only argument.

    Args:
        x: A value that will be returned.

    Returns:
        The argument x.
    '''
    return x



import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestToDictionary(unittest.TestCase):

    def test_to_dictionary(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_dictionary()
        c = {'Aardvark': 'Aardvark',
             'Balloon': 'Balloon',
             'Carrot': 'Carrot',
             'Daisy': 'Daisy',
             'Ecological': 'Ecological'}
        self.assertEqual(b, c)

    def test_to_dictionary_key_selector(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_dictionary(lambda x: x[0])
        c = {'A': 'Aardvark',
             'B': 'Balloon',
             'C': 'Carrot',
             'D': 'Daisy',
             'E': 'Ecological'}
        self.assertEqual(b, c)

    def test_to_dictionary_key_selector_not_callable(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        self.assertRaises(TypeError, lambda: Queryable(a).to_dictionary("not callable"))

    def test_to_dictionary_value_selector(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_dictionary(value_selector=len)
        c = {'Aardvark': 8,
             'Balloon': 7,
             'Carrot': 6,
             'Daisy': 5,
             'Ecological': 10}
        self.assertEqual(b, c)

    def test_to_dictionary_value_selector_not_callable(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        self.assertRaises(TypeError, lambda:Queryable(a).to_dictionary(value_selector="not callable"))

    def test_to_dictionary_duplicate_keys(self):
        a = ['Aardvark', 'Balloon', 'Baboon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_dictionary(lambda x: x[0])
        c = {'A': 'Aardvark',
             'B': 'Baboon',
             'C': 'Carrot',
             'D': 'Daisy',
             'E': 'Ecological'}
        self.assertEqual(b, c)

    def test_to_dictionary_closed(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.to_dictionary())





import unittest
from asq.queryables import Queryable
from helpers import infinite, TracingGenerator

__author__ = "Sixty North"

class TestSelectManyWithIndex(unittest.TestCase):

    def test_select_many_with_index_collection_selector(self):
        a = [{'name' : 'Alice', 'flowers' : ['Agapanthus', 'Allium', 'Alpina', 'Alstroemeria', 'Amaranthus', 'Amarylis' ] },
             {'name' : 'Bob',   'flowers' : ['Bouvardia' ]},
             {'name' : 'Chris', 'flowers' : ['Carnations', 'Cattleya', 'Celosia', 'Chincherinchee', 'Chrysanthemum']}]
        b = Queryable(a).select_many_with_index(lambda i, x: [str(i) + flower for flower in x['flowers']]).to_list()
        c = ['0Agapanthus', '0Allium', '0Alpina', '0Alstroemeria', '0Amaranthus', '0Amarylis', '1Bouvardia',
             '2Carnations', '2Cattleya', '2Celosia', '2Chincherinchee', '2Chrysanthemum']
        self.assertEqual(b, c)

    def test_select_many_with_index_collection_selector_not_callable(self):
        a = [{'name' : 'Alice', 'flowers' : ['Agapanthus', 'Allium', 'Alpina', 'Alstroemeria', 'Amaranthus', 'Amarylis' ] },
             {'name' : 'Bob',   'flowers' : ['Bouvardia' ]},
             {'name' : 'Chris', 'flowers' : ['Carnations', 'Cattleya', 'Celosia', 'Chincherinchee', 'Chrysanthemum']}]
        self.assertRaises(TypeError, lambda: Queryable(a).select_many_with_index("not callable", lambda y: y[:2]))

    def test_select_many_with_index_result_selector_not_callable(self):
        a = [{'name' : 'Alice', 'flowers' : ['Agapanthus', 'Allium', 'Alpina', 'Alstroemeria', 'Amaranthus', 'Amarylis' ] },
             {'name' : 'Bob',   'flowers' : ['Bouvardia' ]},
             {'name' : 'Chris', 'flowers' : ['Carnations', 'Cattleya', 'Celosia', 'Chincherinchee', 'Chrysanthemum']}]
        self.assertRaises(TypeError, lambda: Queryable(a).select_many_with_index(lambda i, x: [str(i) + flower for flower in x['flowers']],
                                                "not callable"))

    def test_select_many_with_index_infinite(self):
        a = infinite()
        b = Queryable(a).select_many_with_index(lambda index, source_element: [source_element] * index).take(10).to_list()
        c = [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]
        self.assertEqual(b, c)

    def test_select_many_with_index_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = Queryable(a).select_many_with_index(lambda index, source_element: [source_element] * index)
        self.assertEqual(a.trace, [])
        b.take(10).to_list()
        self.assertEqual(a.trace, [0, 1, 2, 3, 4])

    def test_select_many_with_index_closed(self):
        a = [{'name' : 'Alice', 'flowers' : ['Agapanthus', 'Allium', 'Alpina', 'Alstroemeria', 'Amaranthus', 'Amarylis' ] },
             {'name' : 'Bob',   'flowers' : ['Bouvardia' ]},
             {'name' : 'Chris', 'flowers' : ['Carnations', 'Cattleya', 'Celosia', 'Chincherinchee', 'Chrysanthemum']}]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.select_many_with_index(lambda i, x: [str(i) + flower for flower in x['flowers']]))


import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestLastOrDefault(unittest.TestCase):

    def test_last_or_default(self):
        a = [42, 45, 23, 12]
        b = Queryable(a).last_or_default(37)
        self.assertEqual(b, 12)

    def test_last_or_default_non_sequence(self):
        def series():
            yield 42
            yield 45
            yield 23
            yield 12
        a = series()
        b = Queryable(a).last_or_default(37)
        self.assertEqual(b, 12)

    def test_last_or_default_empty(self):
        a = []
        b = Queryable(a).last_or_default(37)
        self.assertEqual(b, 37)

    def test_last_or_default_non_sequence_empty(self):
        def series():
            if False:
                yield 42
                yield 45
                yield 23
                yield 12
        a = series()
        b = Queryable(a).last_or_default(37)
        self.assertEqual(b, 37)

    def test_last_or_default_predicate(self):
        a = [37, 54, 57, 23, 12]
        b = Queryable(a).last_or_default(12, lambda x: x >= 50)
        self.assertEqual(b, 57)

    def test_last_or_default_non_sequence_predicate(self):
        def series():
            yield 42
            yield 45
            yield 23
            yield 12
        a = series()
        b = Queryable(a).last_or_default(37, lambda x: x > 15)
        self.assertEqual(b, 23)

    def test_last_or_default_predicate_empty(self):
        a = []
        b = Queryable(a).last_or_default(12, lambda x: x >= 50)
        self.assertEqual(b, 12)

    def test_last_or_default_non_sequence_predicate_empty(self):
        def series():
            if False:
                yield 42
                yield 45
                yield 23
                yield 12
        a = series()
        b = Queryable(a).last_or_default(37, lambda x: x > 15)
        self.assertEqual(b, 37)

    def test_last_or_default_predicate_missing(self):
        a = [37, 42, 23, 12]
        b = Queryable(a).last_or_default(78, lambda x: x >= 50)
        self.assertEqual(b, 78)

    def test_last_or_default_predicate_not_callable(self):
        a = [37, 54, 57, 23, 12]
        self.assertRaises(TypeError, lambda: Queryable(a).last_or_default(12, "not callable"))

    def test_last_or_default_closed(self):
        b = Queryable([])
        b.close()
        self.assertRaises(ValueError, lambda: b.last_or_default(37))



import unittest
import pickle

from asq.record import Record, new

__author__ = "Sixty North"

class TestRecord(unittest.TestCase):

    def test_create(self):
        r = Record(x=10, y=20, z=30)
        self.assertEqual(r.x, 10)
        self.assertEqual(r.y, 20)
        self.assertEqual(r.z, 30)

    def test_create_positional_error(self):
        self.assertRaises(TypeError, lambda: Record(10, 20, 30))

    def test_create_empty(self):
        r = Record()

    def test_record_equality_positive(self):
        a = Record(x=10, y=20, z=30)
        b = Record(x=10, y=20, z=30)
        self.assertTrue(a == b)

    def test_record_equality_negative(self):
        a = Record(x=10, y=20, z=30)
        b = Record(x=10, y=40, z=30)
        self.assertFalse(a == b)

    def test_record_inequality_positive(self):
        a = Record(x=10, y=20, z=30)
        b = Record(x=10, y=40, z=30)
        self.assertTrue(a != b)

    def test_record_inequality_negative(self):
        a = Record(x=10, y=20, z=30)
        b = Record(x=10, y=20, z=30)
        self.assertFalse(a != b)

    def test_record_pickle_roundtrip(self):
        a = Record(x=20, y=80, z=50)
        s = pickle.dumps(a)
        b = pickle.loads(s)
        self.assertEqual(a, b)

    def test_record_str(self):
        a = Record(x=20, y=30, z=40)
        s= str(a)
        self.assertTrue("x=20" in s)
        self.assertTrue("y=30" in s)
        self.assertTrue("z=40" in s)

    def test_record_repr(self):
        a = Record(x=20, y=30, z=40)
        s= repr(a)
        self.assertTrue(s.startswith("Record("))
        self.assertTrue("x=20" in s)
        self.assertTrue("y=30" in s)
        self.assertTrue("z=40" in s)
        self.assertTrue(s.endswith(")"))

    def test_new_create(self):
        r = new(x=10, y=20, z=30)
        self.assertEqual(r.x, 10)
        self.assertEqual(r.y, 20)
        self.assertEqual(r.z, 30)

    def test_new_create_positional_error(self):
        self.assertRaises(TypeError, lambda: new(10, 20, 30))

    def test_new_create_empty(self):
        r = new()











import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestCount(unittest.TestCase):

    def test_count_empty_collection(self):
        b = Queryable([]).count()
        self.assertEqual(b, 0)

    def test_count_empty_sequence(self):
        def empty():
            if False:
                yield 0
        b = Queryable(empty()).count()
        self.assertEqual(b, 0)

    def test_count_finite_collection(self):
        a = [1, 2, 3]
        b = Queryable(a).count()
        self.assertEqual(b, 3)

    def test_count_finite_sequence(self):
        def seq():
            yield 1
            yield 2
            yield 3
        b = Queryable(seq()).count()
        self.assertEqual(b, 3)

    def test_count_predicate(self):
        a = [1, 78, 45, 34, 98, 54, 53]
        b = Queryable(a).count(lambda x: x > 50)
        self.assertEqual(b, 4)

    def test_count_predicate_non_callable(self):
        a = [1, 78, 45, 34, 98, 54, 53]
        self.assertRaises(TypeError, lambda: Queryable(a).count("not callable"))

    def test_count_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.count())


import unittest
from asq.queryables import Queryable
from helpers import TracingGenerator, infinite

__author__ = "Sixty North"

class TestZip(unittest.TestCase):

    def test_zip(self):
        a = [1, 2, 3]
        b = [4, 5, 6]
        c = Queryable(a).zip(b).to_list()
        self.assertEqual(c, [(1, 4), (2, 5), (3, 6)])

    def test_zip_func(self):
        a = [1, 2, 3]
        b = [4, 5, 6]
        c = Queryable(a).zip(b, lambda x, y: int(str(x) + str(y))).to_list()
        self.assertEqual(c, [(14), (25), (36)])

    def test_zip_not_callable(self):
        a = [1, 2, 3]
        b = [4, 5, 6]
        self.assertRaises(TypeError, lambda: Queryable(a).zip(b, "not callable"))

    def test_zip_shorter_longer(self):
        a = [1, 2]
        b = [4, 5, 6]
        c = Queryable(a).zip(b).to_list()
        self.assertEqual(c, [(1, 4), (2, 5)])

    def test_zip_longer_shorter(self):
        a = [1, 2, 3]
        b = [4, 5]
        c = Queryable(a).zip(b).to_list()
        self.assertEqual(c, [(1, 4), (2, 5)])

    def test_zip_empty_longer(self):
        a = []
        b = [4, 5, 6]
        c = Queryable(a).zip(b).to_list()
        self.assertEqual(c, [])

    def test_zip_longer_empty(self):
        a = [1, 2, 3]
        b = []
        c = Queryable(a).zip(b).to_list()
        self.assertEqual(c, [])

    def test_zip_infinite(self):
        c = Queryable(infinite()).zip(infinite()).take(4).to_list()
        self.assertEqual(c, [(0, 0), (1, 1), (2, 2), (3, 3)])

    def test_zip_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = TracingGenerator()
        self.assertEqual(b.trace, [])
        c = Queryable(a).zip(b)
        self.assertEqual(a.trace, [])
        self.assertEqual(b.trace, [])
        d = c.take(4).to_list()
        self.assertEqual(d, [(0, 0), (1, 1), (2, 2), (3, 3)])

    def test_zip_closed(self):
        a = Queryable([1, 2, 3])
        a.close()
        self.assertRaises(ValueError, lambda: a.zip([4, 5, 6]))

    def test_zip_non_iterable(self):
        a = [1, 2, 3]
        b = 5
        self.assertRaises(TypeError, lambda: Queryable(a).zip(b))



import unittest
from asq.queryables import Queryable
from helpers import infinite

__author__ = "Sixty North"

class TestSequenceEqual(unittest.TestCase):

    def test_sequence_equal_positive(self):
        a = [1, 2, 3, 4, 16, 32]
        b = (1, 2, 3, 4, 16, 32)
        c = Queryable(a).sequence_equal(b)
        self.assertTrue(c)

    def test_sequence_equal_negative(self):
        a = [1, 2, 3, 4, 16, 32]
        b = (1, 2, 3, 5, 16, 32)
        c = Queryable(a).sequence_equal(b)
        self.assertFalse(c)

    def test_sequence_equal_shorter_longer(self):
        a = [1, 2, 3]
        b = (1, 2, 3, 4, 16, 32)
        c = Queryable(a).sequence_equal(b)
        self.assertFalse(c)

    def test_sequence_equal_longer_shorter(self):
         a = [1, 2, 3, 4, 5, 6]
         b = (1, 2, 3)
         c = Queryable(a).sequence_equal(b)
         self.assertFalse(c)

    def test_sequence_equal_empty(self):
        a = []
        b = ()
        c = Queryable(a).sequence_equal(b)
        self.assertTrue(c)

    def test_sequence_equal_non_iterable(self):
        a = [1, 2, 3]
        b = None
        self.assertRaises(TypeError, lambda: Queryable(a).sequence_equal(b))

    def test_sequence_equal_comparer(self):
        a = [1, 2, 3, 4, 16, 32]
        b = (-1, 2, -3, 4, -16, 32)
        c = Queryable(a).sequence_equal(b, lambda a, b: abs(a) == abs(b))
        self.assertTrue(c)

    def test_sequence_equal_order(self):
        a = [1, 2]
        b = (2, 1)
        c = Queryable(a).sequence_equal(b)
        self.assertFalse(c)

    def test_sequence_equal_finite_infinite(self):
        a = infinite()
        b = (1, 2, 3, 5, 16, 32)
        c = Queryable(a).sequence_equal(b)
        self.assertFalse(c)

    def test_sequence_equal_infinite_finite(self):
        a = (1, 2, 3, 5, 16, 32)
        b = infinite()
        c = Queryable(a).sequence_equal(b)
        self.assertFalse(c)

    def test_sequence_equal_equality_comparer_positive(self):
        a = [1, 2, -3, 4, 16, 32]
        b = (-1, 2, 3, -4, 16, -32)
        c = Queryable(a).sequence_equal(b, lambda lhs, rhs: abs(lhs) == abs(rhs))
        self.assertTrue(c)

    def test_sequence_equal_equality_comparer_negative(self):
        a = [1, 2, -3, 4, 16, 32]
        b = (-1, 2, 3, -5, 16, -32)
        c = Queryable(a).sequence_equal(b, lambda lhs, rhs: abs(lhs) == abs(rhs))
        self.assertFalse(c)

    def test_sequence_equal_equality_comparer_not_callable(self):
        a = [1, 2, 3, 4, 16, 32]
        b = (1, 2, 3, 4, 16, 32)
        self.assertRaises(TypeError, lambda: Queryable(a).sequence_equal(b, "not callable"))

    def test_sequence_equal_closed(self):
        a = [1, 2, 3, 4, 16, 32]
        b = (1, 2, 3, 4, 16, 32)
        c = Queryable(a)
        c.close()
        self.assertRaises(ValueError, lambda: c.sequence_equal(b))






import unittest
from asq.queryables import Queryable, Grouping

__author__ = "Sixty North"

class TestToLookup(unittest.TestCase):

    def test_to_lookup(self):
        a = ['Agapanthus', 'Allium', 'Alpina', 'Alstroemeria', 'Amaranthus', 'Amarylis', 'Bouvardia', 'Carnations',
             'Cattleya', 'Celosia', 'Chincherinchee', 'Chrysanthemum']
        b = Queryable(a).to_lookup(lambda x: x[0])
        self.assertEqual(len(b), 3)
        g1 = b['A']
        g2 = b['B']
        g3 = b['C']
        self.assertIsInstance(g1, Grouping)
        self.assertIsInstance(g2, Grouping)
        self.assertIsInstance(g3, Grouping)
        self.assertEqual(g1.to_list(), ['Agapanthus', 'Allium', 'Alpina', 'Alstroemeria', 'Amaranthus', 'Amarylis'])
        self.assertEqual(g2.to_list(), ['Bouvardia'])
        self.assertEqual(g3.to_list(), ['Carnations', 'Cattleya', 'Celosia', 'Chincherinchee', 'Chrysanthemum'])

    def test_to_lookup_key_selector(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_lookup(lambda x: x[0])
        self.assertEqual(len(b), 5)
        g1 = b['A']
        g2 = b['B']
        g3 = b['C']
        g4 = b['D']
        g5 = b['E']
        self.assertIsInstance(g1, Grouping)
        self.assertIsInstance(g2, Grouping)
        self.assertIsInstance(g3, Grouping)
        self.assertIsInstance(g4, Grouping)
        self.assertIsInstance(g5, Grouping)
        self.assertEqual(g1.to_list(), ['Aardvark'])
        self.assertEqual(g2.to_list(), ['Balloon'])
        self.assertEqual(g3.to_list(), ['Carrot'])
        self.assertEqual(g4.to_list(), ['Daisy'])
        self.assertEqual(g5.to_list(), ['Ecological'])

    def test_to_lookup_key_selector_not_callable(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        self.assertRaises(TypeError, lambda: Queryable(a).to_lookup("not callable"))

    def test_to_lookup_value_selector(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_lookup(value_selector=len)
        self.assertEqual(len(b), 5)
        g1 = b['Aardvark']
        g2 = b['Balloon']
        g3 = b['Carrot']
        g4 = b['Daisy']
        g5 = b['Ecological']
        self.assertIsInstance(g1, Grouping)
        self.assertIsInstance(g2, Grouping)
        self.assertIsInstance(g3, Grouping)
        self.assertIsInstance(g4, Grouping)
        self.assertIsInstance(g5, Grouping)
        self.assertEqual(g1.to_list(), [8])
        self.assertEqual(g2.to_list(), [7])
        self.assertEqual(g3.to_list(), [6])
        self.assertEqual(g4.to_list(), [5])
        self.assertEqual(g5.to_list(), [10])

    def test_to_lookup_value_selector_not_callable(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        self.assertRaises(TypeError, lambda: Queryable(a).to_lookup(value_selector="not callable"))

    def test_to_lookup_duplicate_keys(self):
        a = ['Aardvark', 'Balloon', 'Baboon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a).to_lookup(lambda x: x[0])
        self.assertEqual(len(b), 5)
        g1 = b['A']
        g2 = b['B']
        g3 = b['C']
        g4 = b['D']
        g5 = b['E']
        self.assertIsInstance(g1, Grouping)
        self.assertIsInstance(g2, Grouping)
        self.assertIsInstance(g3, Grouping)
        self.assertIsInstance(g4, Grouping)
        self.assertIsInstance(g5, Grouping)
        self.assertEqual(g1.to_list(), ['Aardvark'])
        self.assertEqual(g2.to_list(), ['Balloon', 'Baboon'])
        self.assertEqual(g3.to_list(), ['Carrot'])
        self.assertEqual(g4.to_list(), ['Daisy'])
        self.assertEqual(g5.to_list(), ['Ecological'])

    def test_to_lookup_closed(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.to_lookup())



"""This module contains the definition of the IndexedElement type.
"""

from collections import namedtuple

IndexedElement = namedtuple('IndexedElement', ['index', 'value'])

KeyedElement = namedtuple('KeyedElement', ['key', 'value'])




import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestToList(unittest.TestCase):

    def test_to_list(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        b = Queryable(a).to_list()
        self.assertEqual(a, b)

    def test_to_list_identity(self):
        a = [27, 74, 18, 48, 57, 97]
        b = Queryable(a).to_list()
        self.assertTrue(a is b)

    def test_to_list_closed(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.to_list())



import unittest

from asq.namedelements import KeyedElement
from asq.queryables import Queryable
from helpers import infinite, TracingGenerator


class TestSelectWithIndex(unittest.TestCase):

    def test_select_with_corresponding(self):
        a = [27, 74, 18, 48, 57, 97]
        b = Queryable(a).select_with_correspondence(lambda x: x * 2).to_list()
        c = [(x, x * 2) for x in a]
        self.assertListEqual(b, c)

    def test_select_with_corresponding_infinite(self):
        a = infinite()
        b = Queryable(a).select_with_correspondence(lambda x: x * 2).take(4).to_list()
        c = [(x, x * 2) for x in range(4)]
        self.assertListEqual(b, c)

    def test_select_with_corresponding_deferred(self):
        a = TracingGenerator()
        self.assertListEqual(a.trace, [])
        b = Queryable(a).select_with_correspondence(lambda x: x * 2)
        self.assertListEqual(a.trace, [])
        b.take(3).to_list()
        self.assertListEqual(a.trace, list(range(3)))

    def test_select_with_corresponding_default_selector(self):
        a = [27, 74, 18, 48, 57, 97]
        b = Queryable(a).select_with_correspondence(lambda x: x * 2).to_list()
        c = [KeyedElement(x, x*2) for x in a]
        self.assertListEqual(b, c)

    def test_select_with_corresponding_provided_selector(self):
        a = [27, 74, 18, 48, 57, 97]
        b = Queryable(a).select_with_correspondence(lambda x: x * 2, '{}->{}'.format).to_list()
        c = ['{}->{}'.format(x, x*2) for x in a]
        self.assertListEqual(b, c)

    def test_select_with_corresponding_not_callable(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        self.assertRaises(TypeError, lambda: Queryable(a).select_with_correspondence("not callable"))

    def test_select_with_corresponding_closed(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.select_with_correspondence(lambda x: x * 2))



import itertools
import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

def index_by_identity(sequence, obj):
    '''Returns the index of the instance in the sequence. Comparison is by
    object identity rather than object value. Raises ValueError if there is
    no such item.
    '''
    for index, item in enumerate(sequence):
        if item is obj:
            return index
    raise ValueError("index_by_identity(lst, x): x not in lst")

def sgn(i):
    if i < 0: return -1
    elif i > 0: return +1
    return 0

class TestOrderBy(unittest.TestCase):

    def test_order_by(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        b = Queryable(a).order_by().to_list()
        c = [4, 8, 12, 18, 20, 20, 27, 32, 48, 57, 58, 59, 74, 74, 76, 78, 80, 91, 97]
        self.assertEqual(b, c)

    def test_order_by2(self):
        a = [1, 9, 7, 2, 5, 4, 6, 3, 8, 10]
        b = Queryable(a).order_by().to_list()
        c = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.assertEqual(b, c)

    def test_order_by_key(self):
        a = ['Sort', 'words', 'by', 'length']
        b = Queryable(a).order_by(len).to_list()
        c = ['by', 'Sort', 'words', 'length']
        self.assertEqual(b, c)

    def test_order_by_not_callable(self):
        a = ['Sort', 'words', 'by', 'length']
        self.assertRaises(TypeError, lambda: Queryable(a).order_by("not callable"))

    def test_order_by_closed(self):
        a = ['Sort', 'words', 'by', 'length']
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.order_by(len))



    def test_order_by_stability(self):
        a = [1, 2, 3]
        b = [1, 2, 3]
        c = [4, 5, 6]
        d = [4, 5, 6]
        e = [7, 5, 4]
        f = [7, 5, 4]
        self.assertTrue(a is not b)
        self.assertTrue(c is not d)
        self.assertTrue(e is not f)
        s = [a, b, c, d, e, f]
        for pre_perm in itertools.permutations(s):
            pre_index_a = index_by_identity(pre_perm, a)
            pre_index_b = index_by_identity(pre_perm, b)
            pre_index_c = index_by_identity(pre_perm, c)
            pre_index_d = index_by_identity(pre_perm, d)
            pre_index_e = index_by_identity(pre_perm, e)
            pre_index_f = index_by_identity(pre_perm, f)
            pre_order_a_b = sgn(pre_index_a - pre_index_b)
            pre_order_c_d = sgn(pre_index_c - pre_index_d)
            pre_order_e_f = sgn(pre_index_e - pre_index_f)

            post_perm = Queryable(pre_perm).order_by().to_list()

            post_index_a = index_by_identity(post_perm, a)
            post_index_b = index_by_identity(post_perm, b)
            post_index_c = index_by_identity(post_perm, c)
            post_index_d = index_by_identity(post_perm, d)
            post_index_e = index_by_identity(post_perm, e)
            post_index_f = index_by_identity(post_perm, f)
            post_order_a_b = sgn(post_index_a - post_index_b)
            post_order_c_d = sgn(post_index_c - post_index_d)
            post_order_e_f = sgn(post_index_e - post_index_f)

            self.assertEqual(pre_order_a_b, post_order_a_b)
            self.assertEqual(pre_order_c_d, post_order_c_d)
            self.assertEqual(pre_order_e_f, post_order_e_f)


    def test_then_by(self):
        a = ['sort', 'these', 'words', 'by', 'length', 'and', 'then', 'lexicographically']
        b = Queryable(a).order_by(len).then_by().to_list()
        c = ['by', 'and', 'sort', 'then', 'these', 'words', 'length', 'lexicographically']
        self.assertEqual(b, c)

    def test_then_by_key(self):
        a = ['sort', 'using', 'third', 'letter', 'then', 'second']
        b = Queryable(a).order_by(lambda x: x[2]).then_by(lambda y: y[1]).to_list()
        c = ['second', 'then', 'third', 'using', 'sort', 'letter']
        self.assertEqual(b, c)

    def test_then_by_not_callable(self):
        a = ['sort', 'using', 'third', 'letter', 'then', 'second']
        b = Queryable(a).order_by(lambda x: x[2])
        self.assertRaises(TypeError, lambda: b.then_by("not callable"))

    def test_then_by_closed(self):
        a = ['sort', 'using', 'third', 'letter', 'then', 'second']
        b = Queryable(a).order_by(lambda x: x[2])
        b.close()
        self.assertRaises(ValueError, lambda: b.then_by(lambda x: x[1]))

    def test_order_by_descending(self):
        a = [1, 9, 7, 2, 5, 4, 6, 3, 8, 10]
        b = Queryable(a).order_by_descending().to_list()
        c = [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]
        self.assertEqual(b, c)

    def test_order_by_descending_key(self):
        a = ['Sort', 'words', 'by', 'length']
        b = Queryable(a).order_by_descending(len).to_list()
        c = ['length', 'words', 'Sort', 'by']
        self.assertEqual(b, c)

    def test_order_by_descending_not_callable(self):
        a = ['Sort', 'words', 'by', 'length']
        self.assertRaises(TypeError, lambda: Queryable(a).order_by_descending("not callable"))

    def test_order_by_descending_closed(self):
        a = ['Sort', 'words', 'by', 'length']
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.order_by_descending(len))

    def test_then_by_descending(self):
        a = ['sort', 'these', 'words', 'by', 'length', 'and', 'then', 'lexicographically']
        b = Queryable(a).order_by(len).then_by_descending().to_list()
        c = ['by', 'and', 'then', 'sort', 'words', 'these', 'length', 'lexicographically']
        self.assertEqual(b, c)

    def test_then_by_descending_key(self):
        a = ['sort', 'using', 'third', 'letter', 'then', 'second']
        b = Queryable(a).order_by(lambda x: x[2]).then_by_descending(lambda y: y[1]).to_list()
        c = ['second', 'then', 'using', 'third', 'sort', 'letter']
        self.assertEqual(b, c)

    def test_then_by_descending_not_callable(self):
        a = ['sort', 'using', 'third', 'letter', 'then', 'second']
        b = Queryable(a).order_by(lambda x: x[2])
        self.assertRaises(TypeError, lambda: b.then_by_descending("not callable"))

    def test_then_by_descending_closed(self):
        a = ['sort', 'using', 'third', 'letter', 'then', 'second']
        b = Queryable(a).order_by(lambda x: x[2])
        b.close()
        self.assertRaises(ValueError, lambda: b.then_by_descending(lambda x: x[1]))




import unittest
from asq.queryables import Queryable
from helpers import infinite, TracingGenerator

__author__ = "Sixty North"

class TestWhere(unittest.TestCase):

    def test_where(self):
        a = range(0, 100)
        b = Queryable(a).where(lambda x: x % 3 == 0).to_list()
        c = list(range(0, 100, 3))
        self.assertEqual(b, c)

    def test_where_not_callable(self):
        a = range(0, 100)
        self.assertRaises(TypeError, lambda: Queryable(a).where("not callable"))

    def test_where_infinite(self):
        a = infinite()
        b = Queryable(a).where(lambda x: x % 5 == 0).take(3).to_list()
        c = [0, 5, 10]
        self.assertEqual(b, c)

    def test_where_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = Queryable(a).where(lambda x: x % 3 == 0)
        self.assertEqual(a.trace, [])
        c = b.take(2).to_list()
        self.assertEqual(a.trace, [0, 1, 2, 3])


    def test_where_closed(self):
        a = range(0, 100)
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.where(lambda x: x % 3 == 0))


import unittest
from asq.queryables import Queryable
from helpers import infinite

__author__ = "Sixty North"

class TestElementAt(unittest.TestCase):

    def test_element_at(self):
        a = [1, 2, 4, 8, 16, 32, 64, 128]
        b = Queryable(a).element_at(3)
        self.assertEqual(b, 8)

    def test_element_at_out_of_range_infinite(self):
        self.assertRaises(ValueError, lambda: Queryable(infinite()).element_at(-1))

    def test_element_at_out_of_range(self):
        a = [1, 2, 4, 8, 16, 32, 64, 128]
        self.assertRaises(ValueError, lambda: Queryable(a).element_at(20))

    def test_element_at_out_of_range_non_sequence(self):
        def series():
            yield 1
            yield 2
            yield 4
            yield 8
        a = series()
        self.assertRaises(ValueError, lambda: Queryable(a).element_at(20))

    def test_element_at_infinite(self):
        b = Queryable(infinite()).element_at(5)
        self.assertEqual(b, 5)

    def test_element_at_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.element_at(0))



import unittest
from asq.queryables import Queryable
from helpers import infinite, TracingGenerator

__author__ = "Sixty North"

class TestDefaultIfEmpty(unittest.TestCase):

    def test_default_if_empty_empty(self):
        b = Queryable([]).default_if_empty(5).to_list()
        self.assertEqual(b, [5])

    def test_default_if_empty_non_empty(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).default_if_empty(42).to_list()
        self.assertEqual(b, a)

    def test_default_if_empty_infinite(self):
        b = Queryable(infinite()).default_if_empty(42).take(5).to_list()
        self.assertEqual(b, [0, 1, 2, 3, 4])

    def test_default_if_empty_is_deferred_not_empty(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = Queryable(a).default_if_empty(42)
        self.assertEqual(a.trace, [])
        c = b.take(3).to_list()
        self.assertEqual(a.trace, [0, 1, 2])

    def test_default_if_empty_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.default_if_empty(5))



import unittest
from asq.queryables import Queryable
from helpers import infinite

__author__ = "Sixty North"

class TestAny(unittest.TestCase):

    def test_any_empty(self):
        a = []
        b = Queryable(a).any()
        self.assertFalse(b)

    def test_any_not_empty(self):
        a = [False, False, False]
        b = Queryable(a).any()
        self.assertTrue(b)

    def test_any_negative_predicate(self):
        a = [1, 2, 3, 4, 8, 34]
        b = Queryable(a).any(lambda x: x == 15)
        self.assertFalse(b)

    def test_any_positive_predicate(self):
        a = [1, 2, 3, 4, 8, 34]
        b = Queryable(a).any(lambda x: x == 8)
        self.assertTrue(b)

    def test_any_infinite(self):
        b = Queryable(infinite()).any()
        self.assertTrue(b)

    def test_any_infinite_predicate(self):
        b = Queryable(infinite()).any(lambda x: x == 1000)

    def test_any_predicate_not_callable(self):
        self.assertRaises(TypeError, lambda: Queryable([True]).any("not callable"))

    def test_any_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.any())


import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestLast(unittest.TestCase):

    def test_last(self):
        a = [42, 45, 23, 12]
        b = Queryable(a).last()
        self.assertEqual(b, 12)

    def test_last_empty(self):
        self.assertRaises(ValueError, lambda: Queryable([]).last())

    def test_last_non_sequence(self):
        def series():
            yield 42
            yield 45
            yield 23
            yield 12
        a = series()
        b = Queryable(a).last()
        self.assertEqual(b, 12)

    def test_last_predicate_non_sequence(self):
        def series():
            yield 42
            yield 45
            yield 23
            yield 12
        a = series()
        b = Queryable(a).last(lambda x: x > 40)
        self.assertEqual(b, 45)

    def test_empty_non_sequence(self):
        def series():
            if False:
                yield 42
                yield 45
                yield 23
                yield 12
        a = series()
        self.assertRaises(ValueError, lambda: Queryable(a).last())


    def test_empty_non_sequence_predicate(self):
        def series():
            if False:
                yield 42
                yield 45
                yield 23
                yield 12
        a = series()
        self.assertRaises(ValueError, lambda: Queryable(a).last(lambda x: x > 40))


    def test_last_predicate(self):
        a = [37, 54, 57, 23, 12]
        b = Queryable(a).last(lambda x: x >= 50)
        self.assertEqual(b, 57)

    def test_last_predicate_empty(self):
        self.assertRaises(ValueError, lambda: Queryable([]).last(lambda x: x >= 50))

    def test_last_predicate_missing(self):
        a = [37, 42, 23, 12]
        self.assertRaises(ValueError, lambda: Queryable(a).first(lambda x: x >= 50))

    def test_last_predicate_not_callable(self):
        a = [37, 54, 57, 23, 12]
        self.assertRaises(TypeError, lambda: Queryable(a).last("not callable"))

    def test_last_closed(self):
        b = Queryable([])
        b.close()
        self.assertRaises(ValueError, lambda: b.last())




import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"


class TestAverage(unittest.TestCase):

    def test_average(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -8, 3]
        b = Queryable(a).average()
        self.assertEqual(b, 2)

    def test_average_selector(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        b = Queryable(a).average(abs)
        self.assertEqual(b, 5)

    def test_average_empty(self):
        with self.assertRaises(ValueError):
            Queryable([]).average()

    def test_average_selector_non_callable(self):
        a = [5, 7, -3, 2, 1, 9, 3, 2, 1, -15, 7]
        with self.assertRaises(TypeError):
            Queryable(a).average("not callable")

    def test_average_closed(self):
        b = Queryable([1])
        b.close()
        with self.assertRaises(ValueError):
            b.average()




import unittest
from asq.queryables import Queryable

__author__ = "Sixty North"

class TestToStr(unittest.TestCase):

    def test_to_str(self):
        a = "This is a string"
        b = Queryable(a).to_str()
        self.assertEqual(a, b)

    def test_to_str_from_sequence(self):
        a = ["This ", "is ", "a ", "string!"]
        b = "This is a string!"
        c = Queryable(a).to_str()
        self.assertEqual(b, c)

    def test_to_str_from_empty_sequence(self):
        a = []
        b = ""
        c = Queryable(a).to_str()
        self.assertEqual(b, c)

    def test_to_str_empty_with_separator(self):
        a = []
        b = Queryable(a).to_str(separator=', ')
        c = ""
        self.assertEqual(b, c)

    def test_to_str_one_with_separator(self):
        a = ['string']
        b = Queryable(a).to_str(separator=', ')
        c = "string"
        self.assertEqual(b, c)

    def test_to_str_many_with_separator(self):
        a = ['this', 'list', 'will', 'be', 'separated', 'by', 'semicolons']
        b = Queryable(a).to_str(separator='; ')
        c = "this; list; will; be; separated; by; semicolons"
        self.assertEqual(b, c)

    def test_to_str_stringify_separator(self):
        a = ['this', 'list', 'will', 'be', 'separated', 'by', 'fives']
        b = Queryable(a).to_str(separator=5)
        c = "this5list5will5be5separated5by5fives"
        self.assertEqual(b, c)

    def test_stringify_items(self):
        a = [1, 5, 9, 34, 12, 3, 67, 1, 0]
        b = Queryable(a).to_str(separator=', ')
        c = "1, 5, 9, 34, 12, 3, 67, 1, 0"
        self.assertEqual(b, c)

    def test_to_str_closed(self):
        a = ['Aardvark', 'Balloon', 'Carrot', 'Daisy', 'Ecological']
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.to_str())
